/*
 * G-code meta compiler
 *
 * Copyright (C) 2013  B. Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "config.h"

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <string.h>

#include "utils.h"
#include "grammartypes.h"
#include "value.h"
#include "builtin.h"

static const char axes[NAXES] = "XYZABC";

static const char *planename(plane_et p)
{
	switch(p) {
	case PLANE_XY:	return "PLANE_XY";
	case PLANE_XZ:	return "PLANE_XZ";
	case PLANE_YZ:	return "PLANE_YZ";
	}
	return "<unknown plane>";
}

static void check_arg_none(const node_t *ref, const char *id, int argc)
{
	if(argc > 0)
		rtwarning(ref, "%s() has no arguments", id);
}

static void check_arg_one(const node_t *ref, const char *id, int argc)
{
	if(argc < 1)
		rterror(ref, "%s() requires one argument", id);
}

static void check_arg_two(const node_t *ref, const char *id, int argc)
{
	if(argc < 2)
		rterror(ref, "%s() requires two arguments", id);
}

static void check_arg_three(const node_t *ref, const char *id, int argc)
{
	if(argc < 3)
		rterror(ref, "%s() requires three arguments", id);
}

static void check_arg_scalar(const node_t *ref, const char *id, const value_t *arg, int argn)
{
	if(arg->type != VAL_INT && arg->type != VAL_FLOAT)
		rterror(ref, "%s() argument %d must be scalar", id, argn+1);
}

static void check_arg_vector(const node_t *ref, const char *id, const value_t *arg, int argn)
{
	if(arg->type != VAL_VECTOR)
		rterror(ref, "%s() argument %d must be vector", id, argn+1);
}

static void check_arg_unit_none(const node_t *ref, const char *id, const value_t *arg, int argn)
{
	if(arg->unit != UNIT_NONE)
		rtwarning(ref, "%s() argument %d has units attached", id, argn+1);
}

static void check_arg_unit_angle(const node_t *ref, const char *id, const value_t *arg, int argn)
{
	/* XXX: if we decide to make UNIT_DEG and UNIT_RAD */
	if(arg->unit != UNIT_NONE)
		rtwarning(ref, "%s() argument %d has units attached", id, argn+1);
}

static void check_range_double(const node_t *ref, const char *id, int argn, double val, double mini, double maxi)
{
	if(val < mini || val > maxi)
		rtwarning(ref, "%s() argument %d out of range [%f,%f]", id, argn+1, mini, maxi);
}

static int printscalarval(const value_t *v, char tag, int globalref, int relative, int force)
{
	int rv = 0;
	double coord;
	double offs = 0.0;
	double *gpref = NULL;
	if(globalref >= 0 && globalref < NAXES) {
		offs = *global_offs[globalref];
		gpref = global_pos[globalref];
	}
	switch(v->type) {
	default:
		return 0;
	case VAL_INT:
	case VAL_FLOAT:
		coord = value_to_mm(v);
		if(gpref && cl_relative && !relative)
			coord -= *gpref;
		if(force || !relative || !gpref || fabs(coord + offs) >= EPSILON) {
			fprintf(ofp, " %c%.*f", tag, cl_decimals, coord + offs);
			rv = 1;
		}
		if(gpref) {
			if(cl_relative || relative)
				*gpref += coord;
			else
				*gpref = coord;
		}
		return rv;
	}
}

static value_t *movegoto(const node_t *ref, int argc, const char *gcode, const char *fn, int relative)
{
	int i;
	const value_t *arg;
	int nax = 0;
	check_arg_one(ref, fn, argc);
	arg = value_get(argc, 0);
	check_arg_vector(ref, fn, arg, 0);
	fprintf(ofp, "%s", gcode);
	for(i = 0; i < arg->v.n && i < NAXES; i++) {
		if(printscalarval(arg->v.vals[i], axes[i], i, relative, 0))
			nax++;
	}
	/* If we did not get any output on relative movement, force it */
	if(!nax && relative) {
		for(i = 0; i < arg->v.n && i < NAXES; i++) {
			if(printscalarval(arg->v.vals[i], axes[i], i, relative, 1))
				nax++;
		}
	}
	fprintf(ofp, "\n");
	if(!nax)
		rterror(ref, "All coordinates in %s() undefined", fn);
	return value_new(VAL_UNDEF);
}

static value_t *bi_move(const node_t *ref, int argc)
{
	return movegoto(ref, argc, "G1", "move", 0);
}

static value_t *bi_move_r(const node_t *ref, int argc)
{
	return movegoto(ref, argc, "G1", "move_r", 1);
}

static value_t *bi_goto(const node_t *ref, int argc)
{
	return movegoto(ref, argc, "G0", "goto", 0);
}

static value_t *bi_goto_r(const node_t *ref, int argc)
{
	return movegoto(ref, argc, "G0", "goto_r", 1);
}

static value_t *bi_feedmode(const node_t *ref, int argc)
{
	int i = -1;
	const value_t *arg;
	check_arg_one(ref, "feedmode", argc);
	arg = value_get(argc, 0);
	if(isscalar(arg)) {
		i = value_to_int(arg);
	} else if(isstring(arg)) {
		if(!strcasecmp(arg->str, "inverse"))
			i = 0;
		else if(!strcasecmp(arg->str, "upm"))
			i = 1;
		else if(!strcasecmp(arg->str, "upr"))
			i = 2;
	}
	switch(i) {
	case 0: fprintf(ofp, "G93\n"); break;	/* FIXME: G1/G2/G3 should all have F words added */
	case 1: fprintf(ofp, "G94\n"); break;
	case 2:	fprintf(ofp, "G95\n"); break;
	default:
		rterror(ref, "Feed mode must be one of 0 (\"inverse\"), 1 (\"upm\") or 2 (\"upr\")");
	}
	return value_new(VAL_UNDEF);
}

static value_t *bi_feedrate(const node_t *ref, int argc)
{
	const value_t *arg;
	check_arg_one(ref, "feedrate", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "feedrate", arg, 0);
	fprintf(ofp, "F%.*f\n", cl_decimals, value_to_mm(arg));
	return value_new(VAL_UNDEF);
}

static value_t *bi_spindle(const node_t *ref, int argc)
{
	const value_t *arg;
	check_arg_one(ref, "spindle", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "spindle", arg, 0);
	double d = value_to_double(arg);
	if(fabs(d) < EPSILON)
		fprintf(ofp, "M5\n");			/* Stop */
	else if(d < 0.0)
		fprintf(ofp, "M4\nS%.*f\n", cl_decimals, -d);	/* CCW */
	else
		fprintf(ofp, "M3\nS%.*f\n", cl_decimals, d);	/* CW */
	return value_new(VAL_UNDEF);
}

static value_t *bi_spindlespeed(const node_t *ref, int argc)
{
	const value_t *arg;
	check_arg_one(ref, "spindlespeed", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "spindlespeed", arg, 0);
	double d = value_to_mm(arg);
	if(d < 0.0)
		rterror(ref, "Spindle speed must be >= 0");
	fprintf(ofp, "S%.*f\n", cl_decimals, d);
	return value_new(VAL_UNDEF);
}

static value_t *bi_toolchange(const node_t *ref, int argc)
{
	int j = 0;
	const value_t *arg;
	check_arg_one(ref, "toolchange", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "toolchange", arg, 0);
	int i = value_to_int(arg);
	if(i < 0)
		rterror(ref, "Tool index must be >= 0");
	if(argc > 1) {
		arg = value_get(argc, 1);
		j = value_to_boolean(arg);
	}
	fprintf(ofp, "T%d\n", i);
	if(j)
		fprintf(ofp, "G43 H%d\n", i);
	return value_new(VAL_UNDEF);
}

static value_t *bi_coolant(const node_t *ref, int argc)
{
	const value_t *arg;
	check_arg_one(ref, "coolant", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "coolant", arg, 0);
	int i = value_to_int(arg);
	switch(i) {
	case 0: fprintf(ofp, "M9\n"); break;		/* Coolant off */
	case 1: fprintf(ofp, "M7\n"); break;		/* Coolant mist */
	case 2: fprintf(ofp, "M8\n"); break;		/* Coolant flood */
	case 3: fprintf(ofp, "M7\nM8\n"); break;	/* Coolant mist and flood */
	default:
		rterror(ref, "Coolant must be on of: 0 (off), 1 (mist), 2 (flood) or 3 (mist+flood)");
	}
	return value_new(VAL_UNDEF);
}

static value_t *bi_deg2rad(const node_t *ref, int argc)
{
	const value_t *arg;
	check_arg_one(ref, "deg2rad", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "deg2rad", arg, 0);
	check_arg_unit_angle(ref, "deg2rad", arg, 0);
	return value_new_flt(value_to_double(arg) / 180.0 * M_PI, UNIT_NONE);
}

static value_t *bi_rad2deg(const node_t *ref, int argc)
{
	const value_t *arg;
	check_arg_one(ref, "ras2deg", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "rad2deg", arg, 0);
	check_arg_unit_angle(ref, "rad2deg", arg, 0);
	return value_new_flt(value_to_double(arg) / M_PI * 180.0, UNIT_NONE);
}

static value_t *bi_dwell(const node_t *ref, int argc)
{
	const value_t *arg;
	double d;
	check_arg_one(ref, "dwell", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "dwell", arg, 0);
	d = value_to_double(arg);
	if(d < 0.0)
		rterror(ref, "Dwell cannot have negative seconds (%f)", d);
	fprintf(ofp, "G4 P%.*f\n", cl_decimals, d);
	return value_new(VAL_UNDEF);
}

static value_t *bi_plane(const node_t *ref, int argc)
{
	int i = -1;
	const value_t *arg;
	check_arg_one(ref, "plane", argc);
	arg = value_get(argc, 0);
	if(isscalar(arg)) {
		i = value_to_int(arg);
		if(i < 0 || i > 2)
			goto arg_error;
	} else if(isstring(arg)) {
		if(!strcasecmp(arg->str, "xy"))
			i = PLANE_XY;
		else if(!strcasecmp(arg->str, "xz"))
			i = PLANE_XZ;
		else if(!strcasecmp(arg->str, "yz"))
			i = PLANE_YZ;
		else
			goto arg_error;
	} else {
arg_error:
		rterror(ref, "plane() only accepts 0..2 or string XY|XZ|YZ");
	}
	assert(i >= 0 && i <= 2);
	global_plane = i;
	switch(i) {
	case PLANE_XY:	fprintf(ofp, "G17\n"); break;
	case PLANE_XZ:	fprintf(ofp, "G18\n"); break;
	case PLANE_YZ:	fprintf(ofp, "G19\n"); break;
	}
	return value_new(VAL_UNDEF);
}

static value_t *bi_pi(const node_t *ref, int argc)
{
	check_arg_none(ref, "pi", argc);
	value_t *v = value_new(VAL_FLOAT);
	v->d = M_PI;
	return v;
}

static value_t *bi_position(const node_t *ref, int argc)
{
	int i;
	int nax = 3;	/* Default to XYZ */
	if(argc > 1)
		rtwarning(ref, "postion() has only one optional argument");
	if(argc > 0) {
		const value_t *v = value_get(argc, 1);
		if(v->unit != UNIT_NONE)
			rtwarning(ref, "postion() argument has units attached, ignored");
		nax = value_to_int(v);
		if(nax < 1)
			rterror(ref, "postion() argument must be > 0");
		if(nax > NAXES)
			rtwarning(ref, "postion() argument returns maximum %d vector coordinates", NAXES);
	}
	check_arg_none(ref, "position", argc);
	value_t *v = value_new(VAL_VECTOR);
	/* XXX: check if using rotational axes */
	for(i = 0; i < nax && i < NAXES; i++)
		value_vector_add(v, value_new_flt(*global_pos[i], UNIT_MM));
	return v;
}

static value_t *bi_atan(const node_t *ref, int argc)
{
	double d0, d1;
	const value_t *arg0, *arg1;
	value_t *v;
	check_arg_two(ref, "atan", argc);
	arg0 = value_get(argc, 0);
	arg1 = value_get(argc, 1);
	check_arg_scalar(ref, "atan", arg0, 0);
	check_arg_scalar(ref, "atan", arg1, 1);
	check_arg_unit_none(ref, "atan", arg0, 0);
	check_arg_unit_none(ref, "atan", arg1, 1);
	d0 = value_to_double(arg0);
	d1 = value_to_double(arg1);
	v = value_new(VAL_FLOAT);
	v->d = atan2(d0, d1);
	return v;
}

typedef enum {
	FN_sin,
	FN_cos,
	FN_tan,
	FN_asin,
	FN_acos,
	FN_round,
	FN_floor,
	FN_ceil,
	FN_loge,
	FN_log10,
	FN_log2,
	FN_exp,
	FN_sqrt,
} oneargfns_et;

static value_t *oneargdoubles(const node_t *ref, int argc, const char *fn, oneargfns_et idx)
{
	double d;
	const value_t *arg;
	value_t *v;
	check_arg_one(ref, fn, argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, fn, arg, 0);
	switch(idx) {
	case FN_sin:
	case FN_cos:
	case FN_tan:
		check_arg_unit_none(ref, fn, arg, 0);
		break;
	default:
		break;
	}
	d = value_to_double(arg);
	v = value_new(VAL_FLOAT);
	switch(idx) {
	case FN_sin:	v->d = sin(d); break;
	case FN_cos:	v->d = cos(d); break;
	case FN_tan:	v->d = tan(d); break;
	case FN_asin:	check_range_double(ref, fn, 0, d, -1.0, 1.0); v->d = asin(d); break;
	case FN_acos:	check_range_double(ref, fn, 0, d, -1.0, 1.0); v->d = acos(d); break;
	case FN_round:	v->d = round(d); break;
	case FN_floor:	v->d = floor(d); break;
	case FN_ceil:	v->d = ceil(d); break;
	case FN_loge:	check_range_double(ref, fn, 0, d, EPSILON, HUGE_VAL); v->d = log(d); break;
	case FN_log10:	check_range_double(ref, fn, 0, d, EPSILON, HUGE_VAL); v->d = log10(d); break;
	case FN_log2:	check_range_double(ref, fn, 0, d, EPSILON, HUGE_VAL); v->d = log2(d); break;
	case FN_exp:	v->d = exp(d); break;
	case FN_sqrt:	check_range_double(ref, fn, 0, d, 0.0, HUGE_VAL); v->d = sqrt(d); break;
	}
	return v;
}

#define ONEARGFNS(f)	static value_t *bi_ ## f (const node_t *ref, int argc) { return oneargdoubles(ref, argc, #f, FN_ ## f ); }

ONEARGFNS(sin)
ONEARGFNS(cos)
ONEARGFNS(tan)
ONEARGFNS(asin)
ONEARGFNS(acos)
ONEARGFNS(round)
ONEARGFNS(floor)
ONEARGFNS(ceil)
ONEARGFNS(loge)
ONEARGFNS(log10)
ONEARGFNS(log2)
ONEARGFNS(exp)
ONEARGFNS(sqrt)

static value_t *bi_int(const node_t *ref, int argc)
{
	const value_t *arg;
	value_t *v = value_new(VAL_INT);
	check_arg_one(ref, "int", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "int", arg, 0);
	v->i = arg->type == VAL_INT ? arg->i : (int)arg->d;
	v->unit = arg->unit;
	return v;
}

static value_t *bi_float(const node_t *ref, int argc)
{
	const value_t *arg;
	value_t *v = value_new(VAL_FLOAT);
	check_arg_one(ref, "float", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "float", arg, 0);
	v->d = arg->type == VAL_INT ? (double)arg->i : arg->d;
	v->unit = arg->unit;
	return v;
}

static value_t *bi_abs(const node_t *ref, int argc)
{
	const value_t *arg;
	value_t *v;
	check_arg_one(ref, "abs", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "abs", arg, 0);
	v = value_new(arg->type);
	switch(arg->type) {
	case VAL_INT:	v->i = abs(arg->i); break;
	case VAL_FLOAT:	v->d = fabs(arg->d); break;
	default:	rtinternal(ref, "Argument to abs() should be scalar"); break;
	}
	v->unit = arg->unit;
	return v;
}

static value_t *bi_to_mm(const node_t *ref, int argc)
{
	const value_t *arg;
	value_t *v;
	check_arg_one(ref, "to_mm", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "to_mm", arg, 0);
	v = value_new(arg->type);
	switch(arg->unit) {
	case UNIT_NONE:
	case UNIT_MM:
		if(v->type == VAL_INT)
			v->i = arg->i;
		else
			v->d = arg->d;
		break;
	case UNIT_MIL:
		if(v->type == VAL_INT)
			v->d = (double)arg->i * 0.0254;
		else
			v->d = arg->d * 0.0254;
		v->type = VAL_FLOAT;
		break;
	}
	v->unit = UNIT_MM;
	return v;
}

static value_t *bi_to_mil(const node_t *ref, int argc)
{
	const value_t *arg;
	value_t *v;
	check_arg_one(ref, "to_mil", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "to_mil", arg, 0);
	v = value_new(arg->type);
	switch(arg->unit) {
	case UNIT_NONE:
	case UNIT_MIL:
		if(v->type == VAL_INT)
			v->i = arg->i;
		else
			v->d = arg->d;
		break;
	case UNIT_MM:
		if(v->type == VAL_INT)
			v->d = (double)arg->i / 0.0254;
		else
			v->d = arg->d / 0.0254;
		v->type = VAL_FLOAT;
		break;
	}
	v->unit = UNIT_MIL;
	return v;
}

static value_t *bi_to_none(const node_t *ref, int argc)
{
	const value_t *arg;
	value_t *v;
	check_arg_one(ref, "to_none", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "to_none", arg, 0);
	v = value_new(arg->type);
	if(v->type == VAL_INT)
		v->i = arg->i;
	else
		v->d = arg->d;
	v->unit = UNIT_NONE;
	return v;
}

static value_t *bi_count(const node_t *ref, int argc)
{
	const value_t *arg;
	value_t *v;
	check_arg_one(ref, "count", argc);
	arg = value_get(argc, 0);
	v = value_new(VAL_INT);
	v->unit = UNIT_NONE;
	switch(arg->type) {
	case VAL_VECTOR:	v->i = arg->v.n; break;
	case VAL_VECTORLIST:	v->i = arg->vl.n; break;
	case VAL_UNDEF:
	case VAL_INT:
	case VAL_FLOAT:
	case VAL_STRING:
		rterror(ref, "count() can only count vectors and vectorlists");
		break;
	}
	return v;
}

static void printunit(FILE *fp, unit_et u)
{
	switch(u) {
	case UNIT_MM:	fprintf(fp, "mm"); break;
	case UNIT_MIL:	fprintf(fp, "mil"); break;
	default:	break;
	}
}

static void printvector(FILE *fp, const node_t *ref, const value_t *v)
{
	int i;
	assert(v->type == VAL_VECTOR);
	fprintf(fp, "[");
	for(i = 0; i < v->v.n; i++) {
		if(i)
			fprintf(fp, ",");
		switch(v->v.vals[i]->type) {
		case VAL_UNDEF:		fprintf(fp, "-"); break;
		case VAL_INT:		fprintf(fp, "%d", v->v.vals[i]->i); printunit(fp, v->v.vals[i]->unit); break;
		case VAL_FLOAT:		fprintf(fp, "%.*f", cl_decimals, v->v.vals[i]->d); printunit(fp, v->v.vals[i]->unit); break;
		default:		rtinternal(ref, "Invalid value type %d in vector printing", v->v.vals[i]->type);
		}
	}
	fprintf(fp, "]");
}

static value_t *printfunc(const node_t *ref, int argc, const char *pfx, FILE *fp, int forcenl)
{
	int i, j;
	int seennl = 0;

	if(pfx)
		fprintf(fp, "%s:%d:%d: %s: ", ref->filename, ref->linenr, ref->charnr, pfx);

	for(i = 0; i < argc; i++) {
		const value_t *v = value_get(argc, i);
		if(i == argc-1 && v->type == VAL_STRING && v->str[0] && v->str[strlen(v->str)-1] == '\n')
			seennl = 1;
		switch(v->type) {
		case VAL_UNDEF:		fprintf(fp, "<undef>"); break;
		case VAL_INT:		fprintf(fp, "%d", v->i); printunit(fp, v->unit); break;
		case VAL_FLOAT:		fprintf(fp, "%.*f", cl_decimals, v->d); printunit(fp, v->unit); break;
		case VAL_STRING:	fprintf(fp, "%s", v->str); break;
		case VAL_VECTOR:	printvector(fp, ref, v); break;
		case VAL_VECTORLIST:
			fprintf(fp, "{");
			for(j = 0; j < v->vl.n; j++) {
				if(j)
					fprintf(fp, ",");
				printvector(fp, ref, v->vl.vecs[j]);
			}
			fprintf(fp, "}");
			break;
		}
	}

	if(forcenl && !seennl)
		fprintf(fp, "\n");
	return value_new(VAL_UNDEF);
}

static value_t *bi_literal(const node_t *ref, int argc)
{
	return printfunc(ref, argc, NULL, ofp, 0);
}

static value_t *bi_error(const node_t *ref, int argc)
{
	runtimeerrors++;
	return printfunc(ref, argc, "Runtime error()", stderr, 1);
}

static value_t *bi_warning(const node_t *ref, int argc)
{
	return printfunc(ref, argc, "Runtime warning()", stderr, 1);
}

static value_t *bi_message(const node_t *ref, int argc)
{
	return printfunc(ref, argc, "Runtime message()", stderr, 1);
}


typedef enum {
	FN_arc_cw,
	FN_arc_cw_r,
	FN_arc_ccw,
	FN_arc_ccw_r,
} arcfns_et;

static value_t *allarcs(const node_t *ref, int argc, const char *fn, arcfns_et ft, const char *gcode, int relative)
{
	int i;
	const value_t *endp;
	const value_t *rad;
	check_arg_two(ref, fn, argc);
	endp = value_get(argc, 0);	/* Endpoint */
	rad = value_get(argc, 1);	/* Radius */
	check_arg_vector(ref, fn, endp, 0);
	check_arg_scalar(ref, fn, rad, 1);
	double radius = value_to_mm(rad);
	double vecep[3];
	double vecab[3];
	double normal[3] = {0.0, 0.0, 0.0};
	double center[3] = {0.0, 0.0, 0.0};
	double len = 0.0;
	double lambda;
	int turns = 0;

	if(argc > 2) {
		const value_t *t = value_get(argc, 2);
		check_arg_scalar(ref, fn, t, 2);
		check_arg_unit_none(ref, fn, t, 2);
		turns = value_to_int(t);
		if(turns < 1)
			rterror(ref, "%s() Number of turns must be >= 1", fn);
	}

	if(ft == FN_arc_ccw)
		radius *= -1.0;

	/*
	 * S: startpoint
	 * E: endpoint
	 * e: halfway vector SE
	 * n: normal to e
	 * c: centerpoint vector from halway SE to center
	 *
	 * e = 0.5 * (E - S)
	 * c = e + i*n
	 * i^2 = radius^2 - |e|^2
	 *
	 */
	/* endpoint coordinates */
	vecep[0] = endp->v.n < 1 ? 0.0 : value_to_mm(endp->v.vals[0]);
	vecep[1] = endp->v.n < 2 ? 0.0 : value_to_mm(endp->v.vals[1]);
	vecep[2] = endp->v.n < 3 ? 0.0 : value_to_mm(endp->v.vals[2]);

	/* endpoint - position */
	if(relative) {
		vecab[0] = vecep[0];
		vecab[1] = vecep[1];
		vecab[2] = vecep[2];
	} else {
		vecab[0] = vecep[0] - *global_pos[0];
		vecab[1] = vecep[1] - *global_pos[1];
		vecab[2] = vecep[2] - *global_pos[2];
	}

	/* Check endpoint coordinates */
	switch(global_plane) {
	case PLANE_XY:
		if(endp->v.n < 1 || (isundef(endp->v.vals[0]) && endp->v.n >= 2 && isundef(endp->v.vals[1])))
			rterror(ref, "%s() in plane %s is missing both plane coordinates", fn, planename(global_plane));
		len = vecab[0] * vecab[0] + vecab[1] * vecab[1];
		if(radius < 0.0) {
			normal[0] = -0.5 * vecab[1];
			normal[1] = 0.5 * vecab[0];
		} else {
			normal[0] = 0.5 * vecab[1];
			normal[1] = -0.5 * vecab[0];
		}
		lambda = sqrt(len / (radius*radius - normal[0]*normal[0] - normal[1]*normal[1]));
		center[0] = 0.5 * (vecab[0] + lambda * normal[0]);
		center[1] = 0.5 * (vecab[1] + lambda * normal[1]);
		break;
	case PLANE_XZ:
		if(endp->v.n < 1 || (isundef(endp->v.vals[0]) && endp->v.n >= 3 && isundef(endp->v.vals[2])))
			rterror(ref, "%s() in plane %s is missing both plane coordinates", fn, planename(global_plane));
		len = vecab[0] * vecab[0] + vecab[2] * vecab[2];
		if(radius < 0.0) {
			normal[0] = -0.5 * vecab[2];
			normal[2] = 0.5 * vecab[0];
		} else {
			normal[0] = 0.5 * vecab[2];
			normal[2] = -0.5 * vecab[0];
		}
		lambda = sqrt(len / (radius*radius - normal[0]*normal[0] - normal[2]*normal[2]));
		center[0] = 0.5 * (vecab[0] + lambda * normal[0]);
		center[2] = 0.5 * (vecab[2] + lambda * normal[2]);
		break;
	case PLANE_YZ:
		if(endp->v.n < 2 || (isundef(endp->v.vals[1]) && endp->v.n >= 3 && isundef(endp->v.vals[2])))
			rterror(ref, "%s() in plane %s is missing both plane coordinates", fn, planename(global_plane));
		len = vecab[1] * vecab[1] + vecab[2] * vecab[2];
		if(radius < 0.0) {
			normal[1] = -0.5 * vecab[2];
			normal[2] = 0.5 * vecab[1];
		} else {
			normal[1] = 0.5 * vecab[2];
			normal[2] = -0.5 * vecab[1];
		}
		lambda = sqrt(len / (radius*radius - normal[1]*normal[1] - normal[2]*normal[2]));
		center[1] = 0.5 * (vecab[1] + lambda * normal[1]);
		center[2] = 0.5 * (vecab[2] + lambda * normal[2]);
		break;
	}

	if(len / 4.0 > radius * radius)
		rterror(ref, "%s() radius is less than two times start-to-endpoint distance", fn);

	/* This is the endpoint */
	fprintf(ofp, "%s", gcode);
	for(i = 0; i < endp->v.n && i < 3; i++) {
		fprintf(ofp, " %c%.*f", axes[i], cl_decimals, vecep[i] + *global_offs[i] + (relative ? 0.0 : *global_pos[i]));
		if(relative)
			*global_pos[i] += vecep[i];
		else
			*global_pos[i] = vecep[i];
	}

	/* Place the centerpoint coordinates */
	switch(global_plane) {
	case PLANE_XY:	fprintf(ofp, " I%.*f J%.*f", cl_decimals, center[0], cl_decimals, center[1]); break;
	case PLANE_XZ:	fprintf(ofp, " I%.*f K%.*f", cl_decimals, center[0], cl_decimals, center[2]); break;
	case PLANE_YZ:	fprintf(ofp, " J%.*f K%.*f", cl_decimals, center[1], cl_decimals, center[2]); break;
	}

	if(turns)
		fprintf(ofp, " P%d", turns);

	fprintf(ofp, "\n");

	return value_dup(endp);
}

static value_t *bi_arc_cw(const node_t *ref, int argc)
{
	return allarcs(ref, argc, "arc_cw", FN_arc_cw, "G2", 0);
}

static value_t *bi_arc_cw_r(const node_t *ref, int argc)
{
	return allarcs(ref, argc, "arc_cw_r", FN_arc_cw_r, "G2", 1);
}

static value_t *bi_arc_ccw(const node_t *ref, int argc)
{
	return allarcs(ref, argc, "arc_ccw", FN_arc_ccw, "G3", 0);
}

static value_t *bi_arc_ccw_r(const node_t *ref, int argc)
{
	return allarcs(ref, argc, "arc_ccw_r", FN_arc_ccw_r, "G3", 1);
}

static value_t *bi_normalize(const node_t *ref, int argc)
{
	int i;
	const value_t *arg;
	value_t *v;
	double sum = 0.0;
	check_arg_one(ref, "normalize", argc);
	arg = value_get(argc, 0);
	check_arg_vector(ref, "normalize", arg, 0);

	for(i = 0; i < arg->v.n; i++) {
		double d = value_to_mm(arg->v.vals[i]);
		sum += d*d;
	}
	sum = sqrt(sum);
	v = value_new(VAL_VECTOR);
	for(i = 0; i < arg->v.n; i++) {
		switch(arg->v.vals[i]->type) {
		case VAL_UNDEF:
			value_vector_add_undef(v);
			break;
		case VAL_INT:
			value_vector_add(v, value_new_flt(value_to_mm(arg->v.vals[i]) / sum, UNIT_NONE));
			break;
		case VAL_FLOAT:
			value_vector_add(v, value_new_flt(value_to_mm(arg->v.vals[i]) / sum, UNIT_NONE));
			break;
		default:
			rtinternal(ref, "Vector value contains non-scalar elements");
			break;
		}
		
	}
	return v;
}

static value_t *bi_length(const node_t *ref, int argc)
{
	int i;
	const value_t *arg;
	double sum = 0.0;
	unit_et u = UNIT_NONE;
	check_arg_one(ref, "length", argc);
	arg = value_get(argc, 0);
	check_arg_vector(ref, "length", arg, 0);

	for(i = 0; i < arg->v.n; i++) {
		if(arg->v.vals[i]->type == VAL_UNDEF)
			continue;
		assert(isscalar(arg->v.vals[i]));
		if(u == UNIT_NONE)
			u = arg->v.vals[i]->unit;
		else if(u != arg->v.vals[i]->unit) {
			u = -1;
			break;
		}
	}
	for(i = 0; i < arg->v.n; i++) {
		double d;
		switch(u) {
		case UNIT_MM:
		case UNIT_MIL:
		case UNIT_NONE:
			d = value_to_double(arg->v.vals[i]);
			break;
		default:
			d = value_to_mm(arg->v.vals[i]);
		}
		sum += d*d;
	}
	return value_new_flt(sqrt(sum), (-1 == u ? UNIT_MM : u));
}

static value_t *bi_rotate_xy(const node_t *ref, int argc)
{
	int i;
	const value_t *arg0;
	const value_t *arg1;
	double phi, xr, yr;
	value_t *v, *x, *y;
	check_arg_two(ref, "rotate_xy", argc);
	arg0 = value_get(argc, 0);
	arg1 = value_get(argc, 1);
	check_arg_vector(ref, "rotate_xy", arg0, 0);
	check_arg_vector(ref, "rotate_xy", arg1, 1);

	v = value_new(VAL_VECTOR);

	if(arg0->v.n < 2) {
		if(arg0->v.n < 1)
			x = value_new_flt(0.0, UNIT_NONE);
		else
			x = value_new_flt(value_to_double(arg0->v.vals[0]), arg0->v.vals[0]->unit);
		y = value_new_flt(0.0, UNIT_NONE);
	} else {
		x = value_new_flt(value_to_double(arg0->v.vals[0]), arg0->v.vals[0]->unit);
		y = value_new_flt(value_to_double(arg0->v.vals[1]), arg0->v.vals[1]->unit);
	}

	if(x->unit != y->unit) {
		if(x->unit == UNIT_NONE)
			x->unit = y->unit;
		else if(y->unit == UNIT_NONE)
			y->unit = x->unit;
		else {
			x->d = value_to_mm(x);
			y->d = value_to_mm(y);
			x->unit = y->unit = UNIT_MM;
		}
	}

	if(arg1->unit != UNIT_NONE)
		rtwarning(ref, "rotate_xy(): angle has units associated, ignored");
	phi = value_to_double(arg1);

	xr = x->d * cos(phi) - y->d * sin(phi);
	yr = x->d * sin(phi) + y->d * cos(phi);
	y->d = yr;
	x->d = xr;
	value_vector_add(v, x);
	value_vector_add(v, y);
	for(i = 2; i < arg0->v.n; i++)
		value_vector_add(v, value_dup(arg0->v.vals[i]));
	return v;
}

static value_t *bi_rotate_xz(const node_t *ref, int argc)
{
	int i;
	const value_t *arg0;
	const value_t *arg1;
	double phi, xr, zr;
	value_t *v, *x, *z;
	check_arg_two(ref, "rotate_xz", argc);
	arg0 = value_get(argc, 0);
	arg1 = value_get(argc, 1);
	check_arg_vector(ref, "rotate_xz", arg0, 0);
	check_arg_vector(ref, "rotate_xz", arg1, 1);

	v = value_new(VAL_VECTOR);

	if(arg0->v.n < 3) {
		if(arg0->v.n < 1)
			x = value_new_flt(0.0, UNIT_NONE);
		else
			x = value_new_flt(value_to_double(arg0->v.vals[0]), arg0->v.vals[0]->unit);
		z = value_new_flt(0.0, UNIT_NONE);
	} else {
		x = value_new_flt(value_to_double(arg0->v.vals[0]), arg0->v.vals[0]->unit);
		z = value_new_flt(value_to_double(arg0->v.vals[2]), arg0->v.vals[2]->unit);
	}

	if(x->unit != z->unit) {
		if(x->unit == UNIT_NONE)
			x->unit = z->unit;
		else if(z->unit == UNIT_NONE)
			z->unit = x->unit;
		else {
			x->d = value_to_mm(x);
			z->d = value_to_mm(z);
			x->unit = z->unit = UNIT_MM;
		}
	}

	if(arg1->unit != UNIT_NONE)
		rtwarning(ref, "rotate_xy(): angle has units associated, ignored");
	phi = value_to_double(arg1);

	xr = x->d * cos(phi) - z->d * sin(phi);
	zr = x->d * sin(phi) + z->d * cos(phi);
	z->d = zr;
	x->d = xr;
	value_vector_add(v, x);
	if(arg0->v.n >= 2)
		value_vector_add(v, value_dup(arg0->v.vals[1]));
	else
		value_vector_add_undef(v);
	value_vector_add(v, z);
	for(i = 3; i < arg0->v.n; i++)
		value_vector_add(v, value_dup(arg0->v.vals[i]));
	return v;
}

static value_t *bi_rotate_yz(const node_t *ref, int argc)
{
	int i;
	const value_t *arg0;
	const value_t *arg1;
	double phi, yr, zr;
	value_t *v, *y, *z;
	check_arg_two(ref, "rotate_yz", argc);
	arg0 = value_get(argc, 0);
	arg1 = value_get(argc, 1);
	check_arg_vector(ref, "rotate_yz", arg0, 0);
	check_arg_vector(ref, "rotate_yz", arg1, 1);

	v = value_new(VAL_VECTOR);

	if(arg0->v.n < 3) {
		if(arg0->v.n < 2)
			y = value_new_flt(0.0, UNIT_NONE);
		else
			y = value_new_flt(value_to_double(arg0->v.vals[1]), arg0->v.vals[1]->unit);
		z = value_new_flt(0.0, UNIT_NONE);
	} else {
		y = value_new_flt(value_to_double(arg0->v.vals[1]), arg0->v.vals[1]->unit);
		z = value_new_flt(value_to_double(arg0->v.vals[2]), arg0->v.vals[2]->unit);
	}

	if(y->unit != z->unit) {
		if(y->unit == UNIT_NONE)
			y->unit = z->unit;
		else if(z->unit == UNIT_NONE)
			z->unit = y->unit;
		else {
			y->d = value_to_mm(y);
			z->d = value_to_mm(z);
			y->unit = z->unit = UNIT_MM;
		}
	}

	if(arg1->unit != UNIT_NONE)
		rtwarning(ref, "rotate_xy(): angle has units associated, ignored");
	phi = value_to_double(arg1);

	yr = y->d * cos(phi) - z->d * sin(phi);
	zr = y->d * sin(phi) + z->d * cos(phi);
	z->d = zr;
	y->d = yr;
	if(arg0->v.n >= 1)
		value_vector_add(v, value_dup(arg0->v.vals[0]));
	else
		value_vector_add_undef(v);
	value_vector_add(v, y);
	value_vector_add(v, z);
	for(i = 3; i < arg0->v.n; i++)
		value_vector_add(v, value_dup(arg0->v.vals[i]));
	return v;
}

static value_t *bi_drill(const node_t *ref, int argc)
{
	int i;
	const value_t *vec;
	const value_t *retract;
	const value_t *delta;
	int rep = 0;
	check_arg_three(ref, "drill", argc);
	vec = value_get(argc, 0);
	retract = value_get(argc, 1);
	delta = value_get(argc, 2);
	check_arg_vector(ref, "drill", vec, 0);
	check_arg_scalar(ref, "drill", retract, 1);
	check_arg_scalar(ref, "drill", delta, 2);
	if(argc > 3) {
		const value_t *repeat = NULL;
		repeat = value_get(argc, 3);
		check_arg_scalar(ref, "drill", repeat, 3);
		if((rep = value_to_int(repeat)) < 1)
			rterror(ref, "Drill-cycle repeat (fourth argument) must be >0");
	}

	if(vec->v.n < 3 || !isscalar(vec->v.vals[2]))
		rterror(ref, "Drill-cycle must have a scalar Z-coordinate in first argument");

	fprintf(ofp, "G73");
	for(i = 0; i < vec->v.n && i < 3; i++) {
		printscalarval(vec->v.vals[i], axes[i], i, cl_relative, 0);
	}
	printscalarval(retract, 'R', 3, cl_relative, 1);
	printscalarval(delta, 'Q', -1, 0, 0);
	if(rep)
		fprintf(ofp, " L%d", rep);
	fprintf(ofp, "\n");
	return value_new(VAL_UNDEF);
}

typedef enum {
	FN_isfloat,
	FN_isint,
	FN_isscalar,
	FN_isstring,
	FN_isundef,
	FN_isvector,
	FN_isvectorlist,
} boolfunc_et;

#define BOOLFUNC(f)	static value_t *bi_ ## f(const node_t *ref, int argc) { return boolfuncs(ref, argc, #f, FN_ ## f); }

static value_t *boolfuncs(const node_t *ref, int argc, const char *fn, boolfunc_et ft)
{
	const value_t *arg;
	value_t *v = value_new(VAL_INT);
	check_arg_one(ref, fn, argc);
	arg = value_get(argc, 0);
	switch(ft) {
	case FN_isfloat:	v->i = arg->type == VAL_FLOAT; break;
	case FN_isint:		v->i = arg->type == VAL_INT; break;
	case FN_isscalar:	v->i = arg->type == VAL_FLOAT || arg->type == VAL_INT; break;
	case FN_isstring:	v->i = arg->type == VAL_STRING; break;
	case FN_isundef:	v->i = arg->type == VAL_UNDEF; break;
	case FN_isvector:	v->i = arg->type == VAL_VECTOR; break;
	case FN_isvectorlist:	v->i = arg->type == VAL_VECTORLIST; break;
	}
	return v;
}

BOOLFUNC(isfloat)
BOOLFUNC(isint)
BOOLFUNC(isscalar)
BOOLFUNC(isstring)
BOOLFUNC(isundef)
BOOLFUNC(isvector)
BOOLFUNC(isvectorlist)


builtins_t builtins[] = {
	{ "abs",		bi_abs },
	{ "acos",		bi_acos },
	{ "arc_ccw",		bi_arc_ccw },
	{ "arc_ccw_r",		bi_arc_ccw_r },
	{ "arc_cw",		bi_arc_cw },
	{ "arc_cw_r",		bi_arc_cw_r },
	{ "asin",		bi_asin },
	{ "atan",		bi_atan },
	{ "ceil",		bi_ceil },
	{ "coolant",		bi_coolant },
	{ "cos",		bi_cos },
	{ "count",		bi_count },
	{ "deg2rad",		bi_deg2rad },
	{ "drill",		bi_drill },
	{ "dwell",		bi_dwell },
	{ "error",		bi_error },
	{ "exp",		bi_exp },
	{ "feedmode",		bi_feedmode },
	{ "feedrate",		bi_feedrate },
	{ "float",		bi_float },
	{ "floor",		bi_floor },
	{ "goto",		bi_goto },
	{ "goto_r",		bi_goto_r },
	{ "int",		bi_int },
	{ "isfloat",		bi_isfloat },
	{ "isint",		bi_isint },
	{ "isscalar",		bi_isscalar },
	{ "isstring",		bi_isstring },
	{ "isundef",		bi_isundef },
	{ "isvectolistr",	bi_isvectorlist },
	{ "isvector",		bi_isvector },
	{ "length",		bi_length },
	{ "literal",		bi_literal },
	{ "log10",		bi_log10 },
	{ "log2",		bi_log2 },
	{ "loge",		bi_loge },
	{ "message",		bi_message },
	{ "move",		bi_move },
	{ "move_r",		bi_move_r },
	{ "normalize",		bi_normalize },
	{ "pi",			bi_pi },
	{ "plane",		bi_plane },
	{ "position",		bi_position },
	{ "rad2deg",		bi_rad2deg },
	{ "rotate_xy",		bi_rotate_xy },
	{ "rotate_xz",		bi_rotate_xz },
	{ "rotate_yz",		bi_rotate_yz },
	{ "round",		bi_round },
	{ "sin",		bi_sin },
	{ "spindle",		bi_spindle },
	{ "spindlespeed",	bi_spindlespeed },
	{ "sqrt",		bi_sqrt },
	{ "tan",		bi_tan },
	{ "to_mil",		bi_to_mil },
	{ "to_mm",		bi_to_mm },
	{ "to_none",		bi_to_none },
	{ "toolchange",		bi_toolchange },
	{ "warning",		bi_warning },
	//{ "",	bi_ },
};

const int nbuiltins = NELEM(builtins);
