/*
 * G-code meta compiler
 *
 * Copyright (C) 2013  B. Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __GCMC_GRAMMARTYPES_H
#define __GCMC_GRAMMARTYPES_H

#include <stdio.h>
#include <stdarg.h>

#define EPSILON	1e-16	/* This is where doubles lose precision */
#define NAXES	6	/* We support 6 axes by default XYZABC */

typedef enum {
	PLANE_XY,
	PLANE_XZ,
	PLANE_YZ,
} plane_et;

typedef enum {
	NT_INVALID,
	NT_EXPR,
	NT_EXPRLIST,
	NT_IF,
	NT_FOR,
	NT_FOREACH,
	NT_WHILE,
	NT_DOWHILE,
	NT_FUNCTION,
	NT_ARGLIST,
	NT_CALL,
	NT_BREAK,
	NT_CONTINUE,
	NT_RETURN,
	NT_LOCAL,
} node_et;

typedef enum {
	OP_NULL,
	OP_ADD,
	OP_SUB,
	OP_MUL,
	OP_DIV,
	OP_MOD,
	OP_INDEX,
	OP_LOR,
	OP_LAND,
	OP_EQ,
	OP_NE,
	OP_GT,
	OP_LT,
	OP_GE,
	OP_LE,
	OP_NOT,
	OP_INT,
	OP_FLOAT,
	OP_STRING,
	OP_VECTOR,
	OP_VECTORLIST,
	OP_ASSIGN,
	OP_ADDASSIGN,
	OP_SUBASSIGN,
	OP_MULASSIGN,
	OP_DIVASSIGN,
	OP_MODASSIGN,
	OP_INC,
	OP_DEC,
	OP_DEREF,
} op_et;

typedef enum {
	UNIT_NONE,
	UNIT_MM,
	UNIT_MIL,
} unit_et;

struct __node_t;

typedef struct __alist_t {
	int	n;
	int	na;
	char	**ids;
} alist_t;

typedef struct __nlist_t {
	int	n;
	int	na;
	struct __node_t	**nodes;
} nlist_t;

typedef struct __call_t {
	char		*id;
	struct __node_t	*args;
} call_t;

typedef struct __foreach_t {
	struct __node_t	*stmts;
	struct __node_t	*src;
	char		*dst;
} foreach_t;

typedef struct __for_t {
	struct __node_t	*stmts;
	struct __node_t	*init;
	struct __node_t	*cond;
	struct __node_t	*inc;
} for_t;

typedef struct __if_t {
	struct __node_t	*cond;
	struct __node_t	*ifclause;
	struct __node_t	*elseclause;
} if_t;

typedef struct __function_t {
	char		*id;
	struct __node_t	*args;
	struct __node_t	*body;
} function_t;

typedef struct __expr_t {
	struct __node_t	*left;
	struct __node_t	*right;
	op_et		op;		/* Operation */
	union {
		char		*id;		/* Ident ref */
		int		i;		/* integer const */
		double		d;		/* float const */
		nlist_t		nlist;		/* Vector values */
		//struct __node_t	*vec;		/* Vector list */
	};
	unit_et		unit;		/* mm/mil unit */
} expr_t;

typedef struct __node_t {
	int		linenr;
	int		charnr;
	const char	*filename;
	node_et		type;
	struct __node_t	*next;
	struct __node_t	*prev;
	union {
		nlist_t		nlist;
		alist_t		alist;
		expr_t		expr;
		struct __node_t	*eref;
		call_t		call;
		foreach_t	lfe;
		for_t		lfor;
		if_t		cond;
		function_t	func;
	};
} node_t;

int yyerror(const char *fmt, ...) __attribute__((format (printf, 1, 2)));;
int yyfatal(const char *fmt, ...) __attribute__((format (printf, 1, 2)));;
int rterror(const node_t *n, const char *fmt, ...) __attribute__((format (printf, 2, 3)));;
int rtwarning(const node_t *n, const char *fmt, ...) __attribute__((format (printf, 2, 3)));;
int rtinternal(const node_t *n, const char *fmt, ...) __attribute__((format (printf, 2, 3)));;
void node_delete(node_t *n);
void handle_include(const char *fname);
const char *includepath_get(int idx);

extern FILE *yyin;
extern int yydebug;
extern FILE *ofp;
extern int cl_decimals;
extern int cl_relative;
extern int runtimeerrors;

extern double *global_offs[NAXES];
extern double *global_pos[NAXES];
extern plane_et global_plane;

extern int linenr;
extern int charnr;
extern const char *filename;

extern node_t *scripthead;

#endif
