/*
 * G-code meta compiler
 *
 * Copyright (C) 2013  B. Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
%{
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "utils.h"
#include "grammartypes.h"
#include "builtin.h"

#define myassert(x)	do { \
				if(!(x)) { \
					yyfatal("Assertion failed:(%s:%d): %s", __FILE__, __LINE__, #x); \
				} \
			} while(0);
int yylex(void);

static char *funcname;

static node_t *newnode(int type);
static node_t *alist_new(char *id);
static node_t *alist_add(node_t *l, char *id);
static node_t *elist_new(node_t *e);
static node_t *elist_add(node_t *l, node_t *e);
static node_t *vlist_new(node_t *v);
static node_t *vlist_add(node_t *l, node_t *v);
static node_t *vec_new(node_t *e, int allownull);
static node_t *vec_add(node_t *v, node_t *e);
static node_t *node_new_assign(char *id, int op, node_t *e);
static node_t *node_new(int op, node_t *e);
static node_t *node_add(node_t *n1, node_t *n2);
static node_t *expr_new(node_t *l, node_t *r, int op);
static node_t *expr_new_call(char *id, node_t *e);
static node_t *expr_new_id(char *id);
static node_t *expr_new_idx(char *id, node_t *e);
static node_t *expr_new_int(int i, int unit);
static node_t *expr_new_flt(double d, int unit);
static node_t *expr_new_str(char *str);
static node_t *pushtag(node_t *n);
static node_t *poptag(void);
static node_t *gethead(node_t *n);
static void checkfuncname(const char *fn);

node_t *scripthead;

%}

%union{
	char		*str;
	double		d;
	int		i;
	node_t		*node;
}

/* Explicit tokens */
%token FUNCTION FOR FOREACH DO WHILE IF ELSE BREAK CONTINUE RETURN INCLUDE LOCAL
%token TOPEN TCLOSE MM MIL
%token <str> IDENT STRING
%token <i> NUMBER
%token <d> FLOAT
%token SUBASSIGN ADDASSIGN MULASSIGN DIVASSIGN MODASSIGN
%left LOR
%left LAND
%left EQ NE LT GT LE GE
%left '%'
%left '+' '-'
%left '*' '/'
%left INC DEC
%left '!' UPM

%type <i> optunit assign
%type <str> function
%type <node> lines line compound optstmts stmt
%type <node> arglist optarglist
%type <node> optargs args
%type <node> expr call anynum
%type <node> nums vector
%type <node> vlist optvlist

%%
file	: /* Empty */		{ ; }
	| lines			{ scripthead = gethead($1); }
	| lines error		{ yyerror("Syntax error"); }
	;

lines	: line			{ $$ = $1; }
	| lines line		{ $$ = node_add($1, $2); }
	;

line	: compound			{ $$ = $1; }
	| INCLUDE '(' STRING ')' ';'	{
			/* Include files is a language construct */
			/* We should get here without a lookahead symbol */
			assert(yychar == YYEMPTY);
			handle_include($3);	/* Setup flex */
			$$ = NULL;
		}
	| INCLUDE '(' STRING ')' error	{ yyerror("';' expected"); }
	| INCLUDE '(' STRING error	{ yyerror("')' expected"); }
	| INCLUDE '(' error		{ yyerror("String with include filename expected"); }
	| INCLUDE error			{ yyerror("'(' expected"); }
	| stmt ';'			{ $$ = $1; }
	| stmt error			{ yyerror("';' expected"); }
	| LOCAL arglist ';'		{ $$ = node_new(NT_LOCAL, $2); }
	;

optarglist: /* Empty */		{ $$ = NULL; }
	| arglist		{ $$ = $1; }
	;

arglist	: IDENT			{ $$ = alist_new($1); }
	| arglist ',' IDENT	{ $$ = alist_add($1, $3); }
	| arglist ',' error	{ yyerror("Identifier expected"); }
	;

optstmts: /* Empty */		{ $$ = NULL; }
	| lines			{ $$ = gethead($1); }
	;

stmt	: IDENT assign expr		{ $$ = node_new_assign($1, $2, $3); }
	| IDENT assign error		{ yyerror("Expression expected"); }
	| call				{ $$ = $1; }
	| IDENT INC			{ $$ = node_new_assign($1, OP_INC, NULL); }
	| IDENT DEC			{ $$ = node_new_assign($1, OP_DEC, NULL); }
	| IDENT error			{ yyerror("Assignment, increment or decrement expected"); }
	;

assign	: '='		{ $$ = OP_ASSIGN; }
	| ADDASSIGN	{ $$ = OP_ADDASSIGN; }
	| SUBASSIGN	{ $$ = OP_SUBASSIGN; }
	| MULASSIGN	{ $$ = OP_MULASSIGN; }
	| DIVASSIGN	{ $$ = OP_DIVASSIGN; }
	| MODASSIGN	{ $$ = OP_MODASSIGN; }
	;

foreach	: FOREACH	{ pushtag(newnode(NT_FOREACH)); }
	;

for	: FOR		{ pushtag(newnode(NT_FOR)); }
	;

while	: WHILE		{ pushtag(newnode(NT_WHILE)); }
	;

do	: DO		{ pushtag(newnode(NT_DOWHILE)); }
	;

function: FUNCTION IDENT	{
			if(funcname)
				yyerror("Function in functions not supported");
			checkfuncname($2);
			$$ = funcname = $2;
		}
	| FUNCTION error	{ yyerror("Function name expected"); }
	;

compound: foreach '(' expr ';' IDENT ')' TOPEN optstmts TCLOSE		{
			$$ = poptag();
			$$->lfe.src = $3;
			$$->lfe.dst = $5;
			$$->lfe.stmts = $8;
		}
	| foreach '(' expr ';' IDENT ')' TOPEN optstmts error		{ yyerror("'}' expected"); }
	| foreach '(' expr ';' IDENT ')' error				{ yyerror("'{' expected"); }
	| foreach '(' expr ';' IDENT error				{ yyerror("')' expected"); }
	| foreach '(' expr ';' error					{ yyerror("Identifier expected"); }
	| foreach '(' expr error					{ yyerror("';' expected"); }
	| foreach '(' error						{ yyerror("Vectorlist expression expected"); }
	| foreach error							{ yyerror("'(' expected"); }
	| for '(' stmt ';' expr ';' stmt ')' TOPEN optstmts TCLOSE	{
			$$ = poptag();
			$$->lfor.init = $3;
			$$->lfor.cond = $5;
			$$->lfor.inc = $7;
			$$->lfor.stmts = $10;
		}
	| for '(' stmt ';' expr ';' stmt ')' TOPEN optstmts error	{ yyerror("'}' expected"); }
	| for '(' stmt ';' expr ';' stmt ')' error			{ yyerror("'{' expected"); }
	| for '(' stmt ';' expr ';' stmt error				{ yyerror("')' expected"); }
	| for '(' stmt ';' expr ';' error				{ yyerror("Increment statement expected"); }
	| for '(' stmt ';' expr error					{ yyerror("';' expected"); }
	| for '(' stmt ';' error					{ yyerror("Conditional expression expected"); }
	| for '(' stmt error						{ yyerror("';' expected"); }
	| for '(' error							{ yyerror("Initialization statement expected"); }
	| for error							{ yyerror("'(' expected"); }
	| while '(' expr ')' TOPEN optstmts TCLOSE			{
			$$ = poptag();
			$$->lfor.cond = $3;
			$$->lfor.stmts = $6;
		}
	| while '(' expr ')' TOPEN optstmts error			{ yyerror("'}' expected"); }
	| while '(' expr ')' error					{ yyerror("'{' expected"); }
	| while '(' expr error						{ yyerror("')' expected"); }
	| while '(' error						{ yyerror("Conditional expression expected"); }
	| while error							{ yyerror("'(' expected"); }
	| do TOPEN optstmts TCLOSE WHILE '(' expr ')' ';'		{
			$$ = poptag();
			$$->lfor.cond = $7;
			$$->lfor.stmts = $3;
		}
	| do TOPEN optstmts TCLOSE WHILE '(' expr ')' error		{ yyerror("';' expected"); }
	| do TOPEN optstmts TCLOSE WHILE '(' expr error			{ yyerror("')' expected"); }
	| do TOPEN optstmts TCLOSE WHILE '(' error			{ yyerror("Conditional expression expected"); }
	| do TOPEN optstmts TCLOSE WHILE error				{ yyerror("'(' expected"); }
	| do TOPEN optstmts TCLOSE error				{ yyerror("'while' expected"); }
	| do TOPEN optstmts error					{ yyerror("'}' expected"); }
	| do error							{ yyerror("'{' expected"); }
	| function '(' optarglist ')' TOPEN optstmts TCLOSE	{
			$$ = newnode(NT_FUNCTION);
			$$->func.id = $1;
			$$->func.args = $3;
			$$->func.body = $6;
			funcname = NULL;
		}
	| function '(' optarglist ')' TOPEN optstmts error		{ yyerror("'}' expected"); funcname = NULL; }
	| function '(' optarglist ')' error				{ yyerror("'{' expected"); funcname = NULL; }
	| function '(' optarglist error					{ yyerror("')' expected"); funcname = NULL; }
	| function error						{ yyerror("'(' expected"); funcname = NULL; }
	| IF '(' expr ')' TOPEN optstmts TCLOSE				{
			$$ = newnode(NT_IF);
			$$->cond.cond = $3;
			$$->cond.ifclause = $6;
		}
	| IF '(' expr ')' TOPEN optstmts TCLOSE ELSE TOPEN optstmts TCLOSE	{
			$$ = newnode(NT_IF);
			$$->cond.cond = $3;
			$$->cond.ifclause = $6;
			$$->cond.elseclause = $10;
		}
	| IF '(' expr ')' TOPEN optstmts TCLOSE ELSE TOPEN optstmts error	{ yyerror("'}' expected"); }
	| IF '(' expr ')' TOPEN optstmts TCLOSE ELSE error			{ yyerror("'{' expected"); }
	| IF '(' expr ')' TOPEN optstmts error					{ yyerror("'}' expected"); }
	| IF '(' expr ')' error		{ yyerror("'{' expected"); }
	| IF '(' expr error		{ yyerror("')' expected"); }
	| IF '(' error			{ yyerror("Conditional expression expected"); }
	| IF error			{ yyerror("'(' expected"); }
	| RETURN expr ';'		{ $$ = node_new(NT_RETURN, $2); }
	| RETURN error			{ yyerror("Expression expected"); }
	| RETURN expr error		{ yyerror("';' expected"); }
	| BREAK ';'			{ $$ = node_new(NT_BREAK, NULL); }
	| BREAK error			{ yyerror("';' expected"); }
	| CONTINUE ';'			{ $$ = node_new(NT_CONTINUE, NULL); }
	| CONTINUE error		{ yyerror("';' expected"); }
	;

call	: IDENT '(' optargs ')'	{ $$ = expr_new_call($1, $3); }
	;

optargs	: /* Empty */		{ $$ = NULL; }
	| args			{ $$ = $1; }
	;

args	: expr			{ $$ = elist_new($1); }
	| args ',' expr		{ $$ = elist_add($1, $3); }
	| args ',' error	{ yyerror("Expression expected"); }
	| args error		{ yyerror("',' expected"); }
	;

expr	: '+' expr %prec UPM	{ $$ = $2; }
	| '-' expr %prec UPM	{ $$ = expr_new($2, expr_new_int(-1, UNIT_NONE), OP_MUL); }
	| '!' expr		{ $$ = expr_new($2, NULL, OP_NOT); }
	| '(' expr ')'		{ $$ = $2; }
	| expr '+' expr		{ $$ = expr_new($1, $3, OP_ADD); }
	| expr '-' expr		{ $$ = expr_new($1, $3, OP_SUB); }
	| expr '*' expr		{ $$ = expr_new($1, $3, OP_MUL); }
	| expr '/' expr		{ $$ = expr_new($1, $3, OP_DIV); }
	| expr '%' expr		{ $$ = expr_new($1, $3, OP_MOD); }
	| expr EQ expr		{ $$ = expr_new($1, $3, OP_EQ); }
	| expr NE expr		{ $$ = expr_new($1, $3, OP_NE); }
	| expr GT expr		{ $$ = expr_new($1, $3, OP_GT); }
	| expr LT expr		{ $$ = expr_new($1, $3, OP_LT); }
	| expr GE expr		{ $$ = expr_new($1, $3, OP_GE); }
	| expr LE expr		{ $$ = expr_new($1, $3, OP_LE); }
	| expr LAND expr	{ $$ = expr_new($1, $3, OP_LAND); }
	| expr LOR expr		{ $$ = expr_new($1, $3, OP_LOR); }
	| call			{ $$ = $1; }
	| IDENT '[' expr ']'	{ $$ = expr_new_idx($1, $3); }
	| IDENT			{ $$ = expr_new_id($1); }
	| NUMBER optunit	{ $$ = expr_new_int($1, $2); }
	| FLOAT optunit		{ $$ = expr_new_flt($1, $2); }
	| STRING		{ $$ = expr_new_str($1); }
	| vector		{ $$ = $1; }
	| TOPEN optvlist TCLOSE	{ $$ = $2; }
	;

optunit	: /* Empty */		{ $$ = UNIT_NONE; }
	| MM			{ $$ = UNIT_MM; }
	| MIL			{ $$ = UNIT_MIL; }
	;

optvlist: /* Empty */		{ $$ = vlist_new(NULL); }
	| vlist			{ $$ = $1; }
	;

vlist	: expr			{ $$ = vlist_new($1); }
	| vlist ',' expr	{ $$ = vlist_add($1, $3); }
	;

vector	: '[' ']'		{ $$ = vec_new(NULL, 0); }
	| '[' nums ']'		{ $$ = $2; }
	| '[' nums error	{ $$ = $2; yyerror("']' expected"); }
	;

nums	: anynum		{ $$ = vec_new($1, 1); }
	| nums ',' anynum	{ $$ = vec_add($1, $3); }
	| nums ',' error	{ $$ = $1; yyerror("Scalar expression or '-' expected"); }
	;

anynum	: '-'			{ $$ = NULL; }
	| expr			{ $$ = $1; }
	;

%%
static node_t *newnode(int type)
{
	node_t *n = calloc(1, sizeof(node_t));
	myassert(n != NULL);
	n->type = type;
	n->linenr = linenr;
	n->charnr = charnr;
	n->filename = filename;
	return n;
}

static node_t *newnlist(node_t *n, int i)
{
	n->nlist.na = i;
	n->nlist.nodes = calloc(i, sizeof(n->nlist.nodes[0]));
	myassert(n->nlist.nodes != NULL);
	return n;
}

static node_t *newalist(node_t *n, int i)
{
	n->alist.na = i;
	n->alist.ids = calloc(i, sizeof(n->alist.ids[0]));
	myassert(n->alist.ids != NULL);
	return n;
}

static node_t *alist_new(char *id)
{
	node_t *n = newnode(NT_ARGLIST);
	newalist(n, 4);
	n->alist.n = 1;
	n->alist.ids[0] = id;
	return n;
}

static node_t *alist_add(node_t *l, char *id)
{
	int i;
	myassert(l->type == NT_ARGLIST);
	testalloc((void **)&l->alist.ids, l->alist.n, &l->alist.na, sizeof(l->alist.ids[0]));
	for(i = 0; i < l->nlist.n; i++) {
		if(!strcmp(id, l->alist.ids[i])) {
			yyerror("Argument %d's name '%s' already used in argument %d", l->nlist.n+1, id, i+1);
			return l;
		}
	}
	l->alist.ids[l->nlist.n] = id;
	l->alist.n++;
	return l;
}

static node_t *elist_new(node_t *e)
{
	node_t *n = newnode(NT_EXPRLIST);
	newnlist(n, 4);
	n->nlist.n = 1;
	n->nlist.nodes[0] = e;
	return n;
}

static node_t *elist_add(node_t *l, node_t *e)
{
	myassert(e->type == NT_EXPR || e->type == NT_CALL);
	myassert(l->type == NT_EXPRLIST);
	testalloc((void **)&l->nlist.nodes, l->nlist.n, &l->nlist.na, sizeof(l->nlist.nodes[0]));
	l->nlist.nodes[l->nlist.n] = e;
	l->nlist.n++;
	return l;
}

static node_t *vlist_new(node_t *v)
{
	myassert(v == NULL || v->type == NT_EXPR);
	node_t *n = newnode(NT_EXPR);
	n->expr.op = OP_VECTORLIST;
	n->expr.nlist.na = 4;
	n->expr.nlist.nodes = calloc(4, sizeof(n->expr.nlist.nodes[0]));
	myassert(n->expr.nlist.nodes != NULL);
	if(v) {
		n->expr.nlist.n = 1;
		n->expr.nlist.nodes[0] = v;
	}
	return n;
}

static node_t *vlist_add(node_t *l, node_t *v)
{
	myassert(l->type == NT_EXPR);
	myassert(l->expr.op == OP_VECTORLIST);
	myassert(v->type == NT_EXPR);
	testalloc((void **)&l->expr.nlist.nodes, l->expr.nlist.n, &l->expr.nlist.na, sizeof(l->expr.nlist.nodes[0]));
	l->expr.nlist.nodes[l->expr.nlist.n] = v;
	l->expr.nlist.n++;
	return l;
}

static node_t *vec_new(node_t *e, int allownull)
{
	node_t *n = newnode(NT_EXPR);
	n->expr.op = OP_VECTOR;
	n->expr.nlist.na = 4;
	n->expr.nlist.nodes = calloc(4, sizeof(n->expr.nlist.nodes[0]));
	myassert(n->expr.nlist.nodes != NULL);
	if(allownull || e) {
		n->expr.nlist.n = 1;
		n->expr.nlist.nodes[0] = e;
	}
	return n;
}

static node_t *vec_add(node_t *v, node_t *e)
{
	myassert(v->type == NT_EXPR);
	myassert(v->expr.op == OP_VECTOR);
	testalloc((void **)&v->expr.nlist.nodes, v->expr.nlist.n, &v->expr.nlist.na, sizeof(v->expr.nlist.nodes[0]));
	v->expr.nlist.nodes[v->expr.nlist.n] = e;
	v->expr.nlist.n++;
	return v;
}

static node_t *node_new_assign(char *id, int op, node_t *e)
{
	node_t *n = newnode(NT_EXPR);
	n->expr.id = id;
	n->expr.op = op;
	n->expr.left = e;
	return n;
}

static node_t *node_new(int nt, node_t *e)
{
	node_t *n = newnode(nt);
	n->eref = e;
	return n;
}

static node_t *node_add(node_t *n1, node_t *n2)
{
	if(!n1)
		return n2;
	if(!n2)
		return n1;
	n1->next = n2;
	n2->prev = n1;
	return n2;
}

static node_t *expr_new(node_t *l, node_t *r, int op)
{
	node_t *n = newnode(NT_EXPR);
	myassert(l != NULL);
	myassert(r != NULL);
	n->expr.op = op;
	n->expr.left = l;
	n->expr.right = r;
	return n;
}

static node_t *expr_new_call(char *id, node_t *e)
{
	node_t *n = newnode(NT_CALL);
	myassert(e == NULL || e->type == NT_EXPRLIST);
	n->call.id = id;
	n->call.args = e;
	return n;
}

static node_t *expr_new_id(char *id)
{
	node_t *n = newnode(NT_EXPR);
	n->expr.op = OP_DEREF;
	n->expr.id = id;
	return n;
}

static node_t *expr_new_idx(char *id, node_t *e)
{
	node_t *n = newnode(NT_EXPR);
	n->expr.op = OP_INDEX;
	n->expr.id = id;
	n->expr.left = e;
	return n;
}

static node_t *expr_new_int(int i, int unit)
{
	node_t *n = newnode(NT_EXPR);
	n->expr.op = OP_INT;
	n->expr.i = i;
	n->expr.unit = unit;
	return n;
}

static node_t *expr_new_flt(double d, int unit)
{
	node_t *n = newnode(NT_EXPR);
	n->expr.op = OP_FLOAT;
	n->expr.d = d;
	n->expr.unit = unit;
	return n;
}

static node_t *expr_new_str(char *str)
{
	node_t *n = newnode(NT_EXPR);
	n->expr.op = OP_STRING;
	n->expr.id = str;
	return n;
}

static node_t **nodestack;
static int nnodestack;
static int nanodestack;

static node_t *pushtag(node_t *n)
{
	if(!nodestack) {
		nodestack = calloc(16, sizeof(*nodestack));
		assert(nodestack != NULL);
		nnodestack = 0;
		nanodestack = 16;
	} else if(nnodestack >= nanodestack) {
		nodestack = realloc(nodestack, nanodestack * 2 * sizeof(*nodestack));
		assert(nodestack != NULL);
		nanodestack *= 2;
	}
	nodestack[nnodestack] = n;
	nnodestack++;
	return n;
}

static node_t *poptag(void)
{
	assert(nnodestack > 0);
	nnodestack--;
	return nodestack[nnodestack];
}

static node_t *gethead(node_t *n)
{
	if(!n)
		return NULL;
	while(n->prev)
		n = n->prev;
	return n;
}

void node_delete(node_t *head)
{
	node_t *n, *next;
	int i;

	if(nodestack) {
		free(nodestack);
		nodestack = NULL;
	}

	if(!head)
		return;

	for(n = head; n; n = next) {
		next = n->next;
		switch(n->type) {
		case NT_EXPRLIST:
			for(i = 0; i < n->nlist.n; i++)
				node_delete(n->nlist.nodes[i]);
			free(n->nlist.nodes);
			break;
		case NT_ARGLIST:
			for(i = 0; i < n->alist.n; i++)
				free(n->alist.ids[i]);
			free(n->alist.ids);
			break;
		case NT_RETURN:
		case NT_LOCAL:
			node_delete(n->eref);
			break;
		case NT_BREAK:
		case NT_CONTINUE:
		case NT_INVALID:
			break;
		case NT_EXPR:
			node_delete(n->expr.left);
			node_delete(n->expr.right);
			switch(n->expr.op) {
			case OP_NULL:
				break;
			case OP_ADD:
			case OP_SUB:
			case OP_MUL:
			case OP_DIV:
			case OP_MOD:
			case OP_LOR:
			case OP_LAND:
			case OP_EQ:
			case OP_NE:
			case OP_GT:
			case OP_LT:
			case OP_GE:
			case OP_LE:
			case OP_NOT:
			case OP_INT:
			case OP_FLOAT:
				break;

			case OP_INDEX:
			case OP_STRING:
			case OP_ASSIGN:
			case OP_ADDASSIGN:
			case OP_SUBASSIGN:
			case OP_MULASSIGN:
			case OP_DIVASSIGN:
			case OP_MODASSIGN:
			case OP_INC:
			case OP_DEC:
			case OP_DEREF:
				free(n->expr.id);
				break;
			case OP_VECTOR:
			case OP_VECTORLIST:
				for(i = 0; i < n->expr.nlist.n; i++)
					node_delete(n->expr.nlist.nodes[i]);
				free(n->expr.nlist.nodes);
				break;
			}
			break;
		case NT_IF:
			node_delete(n->cond.cond);
			node_delete(n->cond.ifclause);
			node_delete(n->cond.elseclause);
			break;
		case NT_FOR:
		case NT_WHILE:
		case NT_DOWHILE:
			node_delete(n->lfor.stmts);
			node_delete(n->lfor.init);
			node_delete(n->lfor.cond);
			node_delete(n->lfor.inc);
			break;
		case NT_FOREACH:
			node_delete(n->lfe.stmts);
			node_delete(n->lfe.src);
			free(n->lfe.dst);
			break;
		case NT_FUNCTION:
			node_delete(n->func.args);
			node_delete(n->func.body);
			free(n->func.id);
			break;
		case NT_CALL:
			node_delete(n->call.args);
			free(n->call.id);
			break;
		}
		free(n);
	}
}

static const char **funcs;
static int nfuncs;
static int nafuncs;

static void checkfuncname(const char *fn)
{
	int i;
	for(i = 0; i < nbuiltins; i++) {
		if(!strcmp(fn , builtins[i].id))
			yyerror("Function name '%s' reserved as built-in function", fn);
	}
	for(i = 0; i < nfuncs; i++) {
		if(!strcmp(fn , funcs[i]))
			yyerror("Function name '%s' is aready defined", fn);
	}
	testalloc((void **)&funcs, nfuncs, &nafuncs, sizeof(*funcs));
	funcs[nfuncs] = fn;
	nfuncs++;
}
